#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"

static char nesmagic[4] = "NES\032";
static char nsfmagic[5] = "NESM\032";

struct mname
{
    unsigned char num;
    char *name;
};

static struct mname mappernames[] =
{
    {0, "Direct"},
    {1, "MMC1"},
    {2, "UNROM"},
    {3, "CNROM"},
    {4, "MMC3"},
    {5, "MMC5"},
    {7, "AOROM"},
    {9, "MMC2"},
    {24, "VRC6"},
    {64, "RAMBO-1"},
    {66, "GNROM"},
    {68, "Sunsoft #4"},
    {69, "FME-7"},
    {140, "VRC7"},
    {0xFF, "************\nI don't know what mapper this is.\nDon't count on it being supported.\nDon't worry about it, the game probably sucks\n***********\n"}
};

static rominfo ri;
extern int zerofd;
extern int memfd;
int dump_fd;
extern unsigned char dump_audio;

rominfo *loadnsf(char *fname)
{
    FILE *fp = fopen(fname, "rb");
    unsigned char *p;
    
    unsigned long size;
    unsigned long memsize;

    printf("Loading NSF...\n");
    
    fseek(fp, 0, SEEK_END);
    size = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    ri.nsf = 1;
    
    size -= 0x80;
    memsize = size;
    memsize += 4095;
    memsize &= 0xFFFFF000;

    printf("Size: %d, memsize: %d\n", size, memsize);
    p = ri.prgrom = mmap(0, memsize, PROT_READ|PROT_WRITE, MAP_PRIVATE, zerofd, 0);
    if (p == (void *)-1)
	printf("MMAP FAIL\n");
    printf("Zeroing...\n");
    memset(ri.prgrom, 0, memsize);
    printf("Loading header...\n");    
    fseek(fp, 0x6, SEEK_SET);
    fread(&ri.maxsongs, 1, 1, fp);
    fread(&ri.startsong, 1, 1, fp);
    fread(&ri.loadaddr, 2, 1, fp);
    fread(&ri.initaddr, 2, 1, fp);
    fread(&ri.playaddr, 2, 1, fp);
    fseek(fp, 0x6E, SEEK_SET);
    fread(&ri.pbrate, 2, 1, fp);
    fread(ri.switchinit, 8, 1, fp);
    fseek(fp, 0x80, SEEK_SET);
    printf("Loading in at %x\n", ri.loadaddr & 0xFFF);
    fread(&ri.prgrom[ri.loadaddr & 0xFFF], size, 1, fp);
    ri.mapper = 254;

    printf("NSF loaded: INIT %x  LOAD %x  PLAY %x  Maxsongs %x\n",
	   ri.initaddr, ri.loadaddr, ri.playaddr, ri.maxsongs);
    
    fclose(fp);
    
    return &ri;
}

rominfo *loadrom(char *fname)
{
    int i;
    char *sramname;
    FILE *fp;
    char tmpstr[16];
    
    sramname = malloc(strlen(fname)+1);
    strcpy(sramname, fname);

    i = strlen(sramname);
    
    while (sramname[i] != '.')
	i--;
    sramname[i+1] = 0;

    if (dump_audio)
    {
	strcat(sramname,"raw");
	dump_fd = open(sramname, O_CREAT|O_TRUNC|O_NONBLOCK|O_WRONLY);
	sramname[i+1] = 0;
    }
    else
	dump_fd = 0;
    
    strcat(sramname,"srm");

    fp = fopen(fname, "rb");

    if (!fp)
    {
	printf("Could not open rom image %s\n", fname);
	return NULL;
    }

    fread(tmpstr, 1, 16, fp);

    if (strncmp(tmpstr, nesmagic, 4))
    {
	if (strncmp(tmpstr, nsfmagic, 5))
	{
	    printf("Bad magic number\n");
	    return NULL;
	}
	else
	{
	    printf("NSF file...\n");
	    fclose(fp);
	    return loadnsf(fname);
	}
    }
    
    if (tmpstr[8] != 0)
    {
	printf("Bogus data in header, reducing to basic...\n");
	memset(&tmpstr[7], 0, 9);
    }
    
    memset(&ri, 0, sizeof(rominfo));
    
    ri.filename = fname;
    ri.prg_pages = tmpstr[4];
    ri.chr_pages = tmpstr[5];
    ri.mapper = (tmpstr[6]&0xF0) >> 4;
    ri.mapper |= (tmpstr[7]&0xF0);
    ri.f.b = (tmpstr[6] & 0x0F);

    i = 0;

    while (mappernames[i].num != 0xFF)
    {
	if (mappernames[i].num == ri.mapper)
	    break;
	i++;
    }
    if (mappernames[i].num == 0xFF)
    {
	printf("%s", mappernames[i].name);
	return NULL;
    }

    printf("Loading rom: %d PRG pages - %d CHR pages - Mapper %s %s %s\n",
	   ri.prg_pages, ri.chr_pages, mappernames[i].name,
	   ri.f.f.trainer ? "- Trainer" : "", ri.f.f.sram ? "- SRAM" : "");
    
    if (ri.f.f.trainer)
    {
	ri.trainer = malloc(512);
	fread(ri.trainer, 512, 1, fp);
    }
    
    ri.prgrom = mmap(0, 0x4000 * (ri.prg_pages + 1), PROT_READ|PROT_WRITE, MAP_PRIVATE, zerofd, 0);
    ri.prgrom = (unsigned long)ri.prgrom + 0x4000;
    
    printf("prgrom: %p\n", ri.prgrom);
    fread(ri.prgrom, 0x4000 * ri.prg_pages, 1, fp);
    
    if (ri.chr_pages)
    {
	ri.chrrom = mmap(0, 0x2000*ri.chr_pages, PROT_READ|PROT_WRITE, MAP_PRIVATE, zerofd, 0);
	fread(ri.chrrom, 0x2000*ri.chr_pages, 1, fp);
    }

    fread(ri.name, 128, 1, fp);
    if (feof(fp))
	sprintf(ri.name, "UNNAMED");

    fclose(fp);

    ri.sram = mmap(0, 0x2000, PROT_READ|PROT_WRITE, MAP_PRIVATE, zerofd, 0);
    
    if (ri.f.f.sram)
    {
	fp = fopen(sramname, "rb");

	if (!fp)
	{
	    printf("Couldn't open SRAM %s, zeroing...\n", sramname);
	    memset(ri.sram, 0, 0x2000);
	}
	else
	{
	    printf("Loading SRAM...\n");
	    fread(ri.sram, 0x2000, 1, fp);
	    fclose(fp);
	}
    }

    printf("Loading of %s complete\n", ri.name);
    
    return &ri;
}
