#include "sound.h"
const int pulse_lengths [32] =
{5, 127, 10,  1, 20,  2, 40,  3,
 80,  4, 30,  5,  7,  6, 13,  7,
  6,  8, 12,  9, 24, 10, 48, 11,
 96, 12, 36, 13,  8, 14, 16, 15};

const int pulse_widths [4] =
{
  1, //  .......|.......|
  2, //  ...|...|...|...|
  4, //  ..||..||..||..||
  6  //  .|||.|||.|||.|||
};

const int num_channels = 1;

static int sound_length [2];// Length of Sample
static int channels [2];    // Which channels are on?
static int sound_cont [2];  // Sound continuous?
static int pulse_width [2];
static int sound_gmode [2]; // Set Envelope Gain Direct Mode
static int sound_eadat [2]; // Set Evelope/Amplitude Data
static int sweep_on    [2]; // Set Frequency Sweep Toggle
static int sweep_speed [2]; // Frequency Change Speed
static int sweep_method[2]; // Frequency Method 1 = Low-High 0 = High-Low
static int sweep_range [2]; // Frequency Range (0-7)
static int sound_freq  [2]; // Sound Frequency
static int restart [2];

static SAMPLE *buffer;
static byte *sndbuff;
static int voice;


/*
Waveform Types (Channels)
-------------------------

  Pulse (Square)
  --------------
    The NES contains two (2) Pulse channels, which are independant of one
    another.

    The NES lets you, like an analogue synth, change the width of the
    pulse wave (in percentage). This percentage specifies the ratio of
    the On portion ("top") to the Off portion ("bottom").

    The following two (2) diagrams show the difference between two diff-
    erent pulse widths:

      50.0% pulse width                      25.0% pulse width
      +------+      +------+  <----ON----->  +---+            +---+
      |      |      |      |                 |   |            |   |
      |      |      |      |                 |   |            |   |
      |      |      |      |                 |   |            |   |
      +      +------+      +  <----OFF---->  +   +------------+   +

    +---------+----------------------------------------------------------+
    |  $4000  | pAPU Pulse #1 Control Register #1 (W)                    |
    |         |                                                          |
    |         | D7-D6: Pulse Width                                       |
    |         |          00 = 12.5%                                      |
    |         |          01 = 25.0%                                      |
    |         |          10 = 50.0%                                      |
    |         |          11 = 75.0%                                      |
    |         |    D5: Playback Mode                                     |
    |         |           0 = Use Length Counter                         |
    |         |           1 = Continuous (looping)                       |
    |         |    D4: Envelope/Amplitude Gain Direct Mode               |
    |         |           0 = Envelope Mode                              |
    |         |           1 = Amplitude Gain Direct Mode                 |
    |         | D3-D0: Envelope/Amplitude Data                           |
    |         |                                                          |
    |         | See Section 5, Subsection B for more information.        |
    +---------+----------------------------------------------------------+
    |  $4001  | pAPU Pulse #1 Control Register #2 (W)                    |
    |         |                                                          |
    |         |    D7: Frequency Sweeping                                |
    |         |           0 = Off (D6-0 ignored)                         |
    |         |           1 = On  (D6-0 respected)                       |
    |         | D6-D4: Frequency Change Speed                            |
    |         |    D3: Frequency Method                                  |
    |         |           0 = High --> Low                               |
    |         |           1 = Low  --> High                              |
    |         | D2-D0: Frequency Range (0=Min, 7=Max)                    |
    |         |                                                          |
    |         | See Section 5, Subsection F for more information on      |
    |         | frequency sweeping.                                      |
    +---------+----------------------------------------------------------+
    |  $4002  | pAPU Pulse #1 Frequency Register (W16)                   |
    | +$4003  |                                                          |
    |         | D15-D11: Length Data (D7-D3) of $4003 define Length Data |
    |         |  D10-D0: Frequency                                       |
    |         |                                                          |
    |         | See Section 5, Subsection B for a definiton of what the  |
    |         | Length Data specifies.                                   |
    +---------+----------------------------------------------------------+

      Pulse1 = 111860.78 / ($4002 + ($4003 & 7))
      Pulse2 = 111860.78 / ($4006 + ($4007 & 7))
      Tri    = 111860.78 / ($400A + ($400B & 7) / 2)
      Noise  = 111860.78 / ($400E + ($400F & 7))

  Envelope Generator
  ------------------
    The NES has a total of four (4) envelope generators. Two (2) are in-
    dependantly assigned to each of the two (2) Pulse channels, one (1)
    for the Triangle channel, and one (1) for the Noise channel.

    Enveloping consists of four (4) sections: (A)ttack, (D)ecay, (S)ustain,
    and (R)elease; the obvious (and common) abbreviation is ADSR. These sec-
    tions are applied in linear order. Here's a diagram to explain what
    aspect is applied to which portion of a waveform:

      +------+---+----+---+
      |  A   | D | S  | R |
      +------+---+----+---+
      |      |   |    |   |
      |     /\   |    |   |
      |    /  \       |   |
      |   /    \          |
      |  /      \______   |
        /              \ 
       /    WAVEFORM    \ 
      /                  \ 

    ATTACK
    ------
      Attack determines the amount of time a note will take to go from pure
      silence to full volume. A value of zero (0) will cause the note to
      play at full volume immediately, while a larger value will cause the
      note to "fade in."

      Attack on the NES does not seem to be used, henceforth is considered
      at full volume from the start. This applies to all four (4) channels.

    DELAY
    -----
      Delay defines the amount of time to spend lowering the volume between
      the peak volume and Sustain stage. Imagine a piano: when a key is hit,
      the note is originally loud, then decreases in volume, then finally
      sustains until the key is released.

      On the NES, Delay applies to only the Pulse and Triangle channels.

    SUSTAIN
    -------
      Sustain is simply the amount of time to hold a note at a constant
      volume after the Decay stage has completed. A Sustain value equal, or
      greater than that of the Decay value will result in an extremely short
      decay time (virtually null), while a Sustain value of zero (0) will
      not sustain at all, therefore making Sustain totally useless: in this
      circumstance, the Decay value will define how fast the note fades to
      silence.

      On the NES, Sustain applies to all four (4) channels.

    RELEASE
    -------
      Release specifies the amount of time it takes to fade the volume level
      of the note to silence, from the volume level at the Sustain stage. If
      the value of Release is zero (0), the note will IMMEDIATELY terminate
      when the Sustain stage ends (note stops). A higher value will result
      in the note very slowly fading to silence.

      On the NES, Release applies to all four (4) channels.
*/
  void PulseReg1(UINT32 address, UINT8 data, struct MemoryWriteByte *psMemWrite)
  {
    double Pulse1;
    switch (address & 3)
    {
      case 0:

        pulse_width [0] = data >> 6;    // Set Pulse Width
        sound_cont  [0] = data & 0x20;  // Set to continuous if it is
        sound_gmode [0] = data & 0x10;  // Set Envelope Gain Direct Mode
        sound_eadat [0] = data & 0xF;   // Set Evelope/Amplitude Data
        return;
      case 1:

        sweep_on    [0] = data & 0x80;  // Set Frequency Sweep Toggle
        sweep_speed [0] = (data & 0x40) >> 4; // Frequency Change Speed
        sweep_method[0] = data & 0x8;   // Frequency Method 1 = Low-High 0 = High-Low
        sweep_range [0] = data & 7;     // Frequency Range (0-7)
        restart[0] = 1;
        return;

      case 2:
        sound_freq  [0] = data;         // Sound Frequency
          return;
      case 3:
        // Ignoring upper 3 bits right now
        sound_length[0] = pulse_lengths [data >> 3];
        return;

    }
  }

  void PulseReg2(UINT32 address, UINT8 data, struct MemoryWriteByte *psMemWrite)
  {
    //fprintf (dfile, "Access to PulseReg2: addr %x data: %x\n", address, data);
  }
  void TriReg(UINT32 address, UINT8 data, struct MemoryWriteByte *psMemWrite)
  {
    //fprintf (dfile, "Access to TriReg: addr %x data: %x\n", address, data);
  }
  void NoiseReg(UINT32 address, UINT8 data, struct MemoryWriteByte *psMemWrite)
  {
    //fprintf (dfile, "Access to NoiseReg: addr %x data: %x\n", address, data);
  }
  void DeltaReg(UINT32 address, UINT8 data, struct MemoryWriteByte *psMemWrite)
  {
    //fprintf (dfile, "Access to DeltaReg: addr %x data: %x\n", address, data);
  }
  void DMA(UINT32 address, UINT8 data, struct MemoryWriteByte *psMemWrite)
  {
    //fprintf (dfile, "Access to Sprite DMA: addr %x data: %x\n", address, data);
    memcpy(&SRAM[0x0] , &RAM[0x100*data], 256);

  }

  // Register 0x4015 (W)
  void ChanCtrl(UINT32 address, UINT8 data, struct MemoryWriteByte *psMemWrite)
  {
    if (data & 1)
    {
      channels[0] = TRUE; //fprintf (dfile, "Pulse #1 Enabled\n");
      voice_start (voice);
    }
    else
    {
      channels[0] = FALSE; //fprintf (dfile, "Pulse #1 Disabled\n");
      voice_stop (voice);
    }
/*
    if (data & 2)
      fprintf (dfile, "Pulse #2 Enabled\n");
    else
      fprintf (dfile, "Pulse #2 Disabled\n");
    if (data & 4)
      fprintf (dfile, "Triangle Enabled\n");
    else
      fprintf (dfile, "Triangle Disabled\n");
    if (data & 8)
      fprintf (dfile, "Noise Enabled\n");
    else
      fprintf (dfile, "Noise Disabled\n");
    if (data & 0x10)
      fprintf (dfile, "DPCM Enabled\n");
    else
      fprintf (dfile, "DPCM Disabled\n");
    if (data > 0x1F)
      fprintf (dfile, "Hmm...  something wrong here\n");
*/

    //fprintf (dfile, "Access to Channel Control: addr %x data: %x\n", address, data);
  }
  UINT8 DMAStats(UINT32 address, UINT8 data, struct MemoryReadByte *psMemRead)
  {
    //fprintf (dfile, "Access to DMA Stats Read ");
    //m6502zpGetContext(KANE);
    //fprintf (dfile, "PC: %x\n", KANE->m6502pc);
    //fprintf (dfile, "AF: %x  X: %x  Y: %x  S: %x\n",
    //         KANE->m6502af, KANE->m6502x, KANE->m6502y, KANE->m6502s);
    int x;
    int d = 0;
    for (x=0; x < num_channels; x++)
      if (sound_length[x])
        d |= (0xFF & (x+1));
    fprintf (dfile, "Access to DMA Stats Read: %u", d);
      return d;
  }

void UpDateSound (void)
{
  int x;
  int tmp;
  float freq, playratio;
  int xxx, help;
  for (x=0; x < 2; x++) // Do the Pulse Waves first
  {
    if (sound_length[x] || sound_cont[x])
    {
      sound_length[x]--; // Possible Problem
      freq = 111860.78 / sound_freq[0];//(RAM[4002] + (RAM[4003] & 7));
      playratio = 22050/freq;
      fprintf (dfile, "%f     %f\n", freq, playratio);
      for (tmp = 0; tmp < 3680; )
      {
        xxx++;
        if (xxx > 8)
          xxx = 1;
        for (help = playratio; help >= 0; help--)

        if (xxx <= pulse_widths[pulse_width[x]])
          sndbuff[tmp++] = (sound_eadat[0] & 0x0f) * 0x11;
        else
          sndbuff[tmp++] = 0;
      }
      if (restart[0])
      {
        //voice_start (voice);
        play_sample(buffer, 64, 127, NULL, 0);
        restart[0] = 0;
      }
//      while (!key[KEY_F1])
//      {
//      }
/*
const float pulse_widths [4] =
{
  12.5, //  .......|.......|
  25.0, //  ...|...|...|...|
  50.0, //  ..||..||..||..||
  75.0  //  .|||.|||.|||.|||
};
static int sound_length [2];// Length of Sample
static int channels [2];    // Which channels are on?
static int sound_cont [2];  // Sound continuous?
static int pulse_width [2];
static int sound_gmode [2]; // Set Envelope Gain Direct Mode
static int sound_eadat [2]; // Set Evelope/Amplitude Data
static int sweep_on    [2]; // Set Frequency Sweep Toggle
static int sweep_speed [2]; // Frequency Change Speed
static int sweep_method[2]; // Frequency Method 1 = Low-High 0 = High-Low
static int sweep_range [2]; // Frequency Range (0-7)
static int sound_freq  [2]; // Sound Frequency
      //Do sound stuff =(
*/

    }

  }

}

int Init_Sound (void)
{
  if (install_sound (DIGI_SB, DIGI_NONE, NULL))
    return 0xFF;
  fprintf (dfile, "Installed Sound okay\n", allegro_error);
/*
  buffer = load_sample("d:/email.wav");
  if (buffer == NULL)
    return 0xFF;
  fprintf (dfile, "Loaded wavfile okay\n", allegro_error);
  fprintf (dfile, "Errors: %s\n", allegro_error);
*/

  buffer = malloc(sizeof(SAMPLE));  // Create the "SAMPLE"
  buffer->data = malloc(3680);    // 2K should be enough??
  sndbuff = buffer->data;           // A lot easier to type sndbuff then buffer->data :P
  buffer->bits = 8;                 // 8 bits Audio
  buffer->freq = 22050;             // Sample's frequency
//  buffer->freq = 80000;             // Sample's frequency
  buffer->priority = 0xFF;          // Whatever...
  buffer->len = 3680;             // I think this is right
  buffer->param = 0;                // Unsure
  voice = allocate_voice(buffer);
  if (voice > 32)
    return 0xFF;

  return 0;
}

void Free_Sound (void)
{
  FILE *tmp;

  voice_stop (voice);
  deallocate_voice(voice);
  tmp = fopen ("d:/kane.wav", "wb");
  fwrite (buffer->data, 1, 368, tmp);
  free (buffer->data);
  free (buffer);
  remove_sound ();
}


//typedef struct SAMPLE                  /* a sample */
//{
//2 int bits;                           /* 8 or 16 */
//2 int freq;                           /* sample frequency */
//2 int priority;                       /* 0-255 */
//4 unsigned long len;                  /* length (in samples) */
//4 unsigned long loop_start;           /* loop start position */
//4 unsigned long loop_end;             /* loop finish position */
//4 unsigned long param;                /* for internal use by the driver */
//4 void *data;                         /* sample data */
//} SAMPLE;
//malloc(0x10000);
