/*************************************************************************/
/* T6502 -> Tennessee's 6502 CPU core [t6502.c: Emulator CORE]           */
/*                                                                       */
/*************************************************************************/

#include <string.h>

#include "t6502.h"

BYTE tmp_byte1, tmp_byte2; /* Temporary byte values */
tpair tmp_word1, tmp_word2; /* Temporary word values */
tpair effective_address; /* Effective address */
int dist_modifier; /* 1 if distance modifier active */
BYTE opcode; /* Byte containing the current opcode */
CPU_6502_unit *tmp_cpu; /* Temporary CPU structure */
int cycle_modifier; /* Cycle counting modifier */
int carry_save; /* Carry save for ADC and SBC */
/*
static BYTE cycles[256] =
 {0,4,0,0,0,4,6,0,3,4,2,0,0,4,6,0,
  2,4,4,0,0,4,6,0,2,4,0,0,0,4,6,0,
  6,6,0,0,4,4,6,0,4,4,2,0,4,4,6,0,
  2,4,4,0,0,4,6,0,4,4,0,0,0,4,6,0,
  6,4,0,0,0,4,6,0,3,4,2,0,3,4,6,0,
  2,4,4,0,0,4,6,0,2,4,0,0,0,4,6,0,
  6,4,0,0,0,4,6,0,4,4,2,0,3,4,6,0,
  2,4,4,0,0,4,6,0,2,4,0,0,0,4,6,0,
  0,4,0,0,4,4,4,0,2,0,2,0,4,4,4,0,
  2,4,4,0,4,4,4,0,2,4,2,0,0,4,0,0,
  4,4,4,0,4,4,4,0,2,4,2,0,4,4,4,0,
  2,4,4,0,4,4,4,0,2,4,2,0,4,4,4,0,
  4,4,0,0,4,4,6,0,2,4,2,0,4,4,6,0,
  2,4,4,0,0,4,6,0,2,4,0,0,0,4,6,0,
  4,4,0,0,4,4,6,0,2,4,2,0,4,4,6,0,
  2,4,4,0,0,4,6,0,2,4,0,0,0,4,6,0};
*/
/* Macro to set both the Zero and Negative flag if necessary */
/* #define SENZ(p,n) p->status &= ~(FLAG_zero | FLAG_negative); p->status \
 |= (n) ? ((n) & 0x80) ? FLAG_negative : 0 : FLAG_zero
*/
#define SENZ(p,n) if (n) CLEAR_FLAG(p,FLAG_zero); else SET_FLAG(p, FLAG_zero); \
               if (n & 0x80) SET_FLAG(p,FLAG_negative); else CLEAR_FLAG(p, FLAG_negative)

void t6502_init_cpu(CPU_6502_unit *CPU, BYTE (*read_func)(WORD A), BYTE (*read_op_func)(WORD A), void (*write_func)(BYTE d, WORD A), char *base_name ) {
 /* Set basic registers to 0 */
 CPU->A = 0;
 CPU->X = 0;
 CPU->Y = 0;

 /* Set name */
 strncpy(CPU->name, base_name, 16);
 CPU->name[16]='\0';
 
 /* Set flags */
 SET_FLAG(CPU,FLAG_zero | FLAG_unused);

 /* Set the read and write functions */
 CPU->read_mem = read_func; /* Memory read function */
 CPU->read_op = read_op_func; /* Opcode read function */
 CPU->write_mem = write_func; /* Memory write function */

 /* Get entry point PC */
 CPU->PC.byte.l = CPU->read_op(0xFFFC);
 CPU->PC.byte.h = CPU->read_op(0xFFFD);

 /* Set cycles to 0 */
 CPU->cycles = 0;

 CPU->int_nesting = 0;

}

void t6502_trigger_IRQ(CPU_6502_unit *CPU) {
 if (!IS_FLAG_SET(CPU, FLAG_interrupt)) { /* Check if interrupt are disabled */
    CPU->cycles += 7; /* Add 7 cycles for interrput generation */

    /* Push PC */
    PUSH_BYTE(CPU, CPU->PC.byte.h);
    PUSH_BYTE(CPU, CPU->PC.byte.l);

    /* Push status register with B bit set */
    PUSH_BYTE(CPU, CPU->status &~ FLAG_BRK);

    /* Disable further interrupts */
    SET_FLAG(CPU,FLAG_interrupt);


    /* Set the PC to the address stored in the vector */
    CPU-> PC.byte.l=CPU->read_op(0xFFFE);
    CPU-> PC.byte.h=CPU->read_op(0xFFFF);

    CPU->int_nesting++;
 }
}

void t6502_trigger_NMI(CPU_6502_unit *CPU) {
    CPU->cycles += 7; /* Add 7 cycles for interrput generation */

    /* Push PC */
    PUSH_BYTE(CPU, CPU->PC.byte.h);
    PUSH_BYTE(CPU, CPU->PC.byte.l);

    /* Push status register */
    PUSH_BYTE(CPU, CPU->status);

    /* Set the PC to the address stored in the vector */
    CPU-> PC.byte.l=CPU->read_op(0xFFFA);
    CPU-> PC.byte.h=CPU->read_op(0xFFFB);

    CPU->int_nesting++;
}

void t6502_trigger_BRK(CPU_6502_unit *CPU) {
 if (!IS_FLAG_SET(CPU, FLAG_interrupt)) { /* Check if interrupt are disabled */
    CPU->cycles += 7; /* Add 7 cycles for interrput generation */

    CPU->PC.word++;
    /* Push PC */
    PUSH_BYTE(CPU, CPU->PC.byte.h);
    PUSH_BYTE(CPU, CPU->PC.byte.l);

    /* Push status register with B bit set */
    PUSH_BYTE(CPU, CPU->status &~ FLAG_BRK);

    /* Disable further interrupts */
    SET_FLAG(CPU,FLAG_interrupt);


    /* Set the PC to the address stored in the vector */
    CPU-> PC.byte.l=CPU->read_op(0xFFFE);
    CPU-> PC.byte.h=CPU->read_op(0xFFFF);
    CPU->int_nesting++;
 }
}

void t6502_trigger_RESET(CPU_6502_unit *CPU) {
    CPU-> PC.byte.l=CPU->read_op(0xFFFC);
    CPU-> PC.byte.h=CPU->read_op(0xFFFD);
}


/* ADDRESSING MODES */

/* Immediate */
INLINE void IMMED(void) {
 tmp_cpu->PC.word++;
 effective_address.word = tmp_cpu->PC.word;
 cycle_modifier = -2;
}

/* Absolute */
INLINE void ABSOL(void) {
 tmp_cpu->PC.word++;
 effective_address.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 tmp_cpu->PC.word++;
 effective_address.byte.h = tmp_cpu->read_op(tmp_cpu->PC.word);
 cycle_modifier = 0;
}

/* Zero page */
INLINE void ZEROP(void) {
 tmp_cpu->PC.word++;
 effective_address.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 effective_address.byte.h = 0;
 cycle_modifier = -1;
}

/* Implicit */
INLINE void IMPLI(void) {
 cycle_modifier = 0;
}


/* Indirect Absolute */
INLINE void INDIA(void) {
 tmp_cpu->PC.word++;
 tmp_word1.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 tmp_cpu->PC.word++;
 tmp_word1.byte.h = tmp_cpu->read_op(tmp_cpu->PC.word);
 effective_address.byte.l = tmp_cpu->read_op(tmp_word1.word);
 effective_address.byte.h = tmp_cpu->read_op(tmp_word1.word+1);
 cycle_modifier = 2;
}

/* Absolute indexed with X */
INLINE void ABSIX(void) {
 tmp_cpu->PC.word++;
 tmp_word1.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 tmp_cpu->PC.word++;
 tmp_word1.byte.h = tmp_cpu->read_op(tmp_cpu->PC.word);
 effective_address.word = tmp_word1.word + tmp_cpu->X;
 if (effective_address.byte.h != tmp_word1.byte.h)
  cycle_modifier = 1;
 else
  cycle_modifier = 0;
}

/* Absolute indexed with Y */
INLINE void ABSIY(void) {
 tmp_cpu->PC.word++;
 tmp_word1.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 tmp_cpu->PC.word++;
 tmp_word1.byte.h = tmp_cpu->read_op(tmp_cpu->PC.word);
 effective_address.word = tmp_word1.word + tmp_cpu->Y;
 if (effective_address.byte.h != tmp_word1.byte.h) cycle_modifier = 1;
 else
  cycle_modifier = 0;
}

/* Zero Page indexed with X */
INLINE void ZEPIX(void) {
 tmp_cpu->PC.word++;
 effective_address.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 effective_address.byte.h = 0;
 effective_address.word += tmp_cpu->X;
 cycle_modifier = 0;
}

/* Zero Page indexed with Y */
INLINE void ZEPIY(void) {
 tmp_cpu->PC.word++;
 effective_address.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 effective_address.byte.h = 0;
 effective_address.word += tmp_cpu->Y;
 cycle_modifier = 0;
}

/* Indexed Indirect */
INLINE void INDIN(void) {
 tmp_cpu->PC.word++;
 tmp_word1.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 tmp_word1.byte.h = 0;
 tmp_word1.byte.l = (BYTE)(tmp_word1.byte.l + tmp_cpu->X);
 tmp_word2.byte.h = 0;
 tmp_word2.byte.l = (BYTE)(tmp_word1.byte.l + 1);
 effective_address.byte.l = tmp_cpu->read_op(tmp_word1.word);
 effective_address.byte.h = tmp_cpu->read_op(tmp_word2.word);
 cycle_modifier = 2;
}

/* Indirect Indexed */
INLINE void ININD(void) {
 tmp_cpu->PC.word++;
 tmp_word1.byte.l = tmp_cpu->read_op(tmp_cpu->PC.word);
 tmp_word1.byte.h = 0;
 effective_address.byte.l = tmp_cpu->read_op(tmp_word1.word);
 effective_address.byte.h = tmp_cpu->read_op(tmp_word1.word+1);
 tmp_word1.word = effective_address.word;
 effective_address.word += tmp_cpu->Y;
 if (tmp_word1.byte.h != effective_address.byte.h)
  cycle_modifier = 2;
 else
  cycle_modifier = 1;
}

/* Relative */
INLINE void RELAT(void) {
 tmp_cpu->PC.word++;
 tmp_byte1 = tmp_cpu->read_op(tmp_cpu->PC.word);
 effective_address.word = tmp_cpu->PC.word + (toffset)(tmp_byte1) + 1;
 if (tmp_cpu->PC.byte.h != effective_address.byte.h)
  cycle_modifier = 2;
 else
  cycle_modifier = 1;
}

/* ADC -> ADd with Carry */
/**********************************************************************/
INLINE void O_ADC(void) {
 int tmp_word3;

 /* Get value from adress */
 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

 carry_save = IS_FLAG_SET(tmp_cpu,FLAG_carry); /* Save the carry */

 /* Do the addition */
 tmp_word3 = (char)tmp_cpu->A + (char)tmp_byte1 + (BYTE)carry_save;

 /* Check and set overflow */
 if ((tmp_word3 > 127) || (tmp_word3 < -128))
  SET_FLAG(tmp_cpu,FLAG_overflow);
 else
  CLEAR_FLAG(tmp_cpu,FLAG_overflow);
  
 tmp_word3 = (BYTE)tmp_cpu->A + (BYTE)tmp_byte1 + carry_save;

 /* Check for carry */
 if (tmp_word3 > 0xff)
  SET_FLAG(tmp_cpu,FLAG_carry);
 else
  CLEAR_FLAG(tmp_cpu,FLAG_carry);

 /* Put result in accumulator */
 tmp_cpu->A = tmp_word3;

 /* Adjust A if decimal mode is set */
 if (IS_FLAG_SET(tmp_cpu,FLAG_decimal)) {
  CLEAR_FLAG(tmp_cpu,FLAG_carry);

  if ((tmp_cpu->A & 0x0f) > 0x09)
   tmp_cpu->A += 0x06;

  if ((tmp_cpu->A & 0xf0) > 0x90) {
   tmp_cpu->A += 0x60;
   SET_FLAG(tmp_cpu,FLAG_carry);
  }
 }

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* AND -> bitwise AND on memory */

INLINE void O_AND(void) {
 tmp_cpu->A &= tmp_cpu->read_mem(effective_address.word);

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}


/* ASL -> Arithmetic Shift Left */
INLINE void O_ASL(void) {
 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

 CLEAR_FLAG(tmp_cpu,FLAG_carry);
 if (tmp_byte1 & 0x80) SET_FLAG(tmp_cpu,FLAG_carry);

 /* Do the bit shift */
 tmp_byte1 <<= 1;

 /* Write back to memory */
 tmp_cpu->write_mem(tmp_byte1, effective_address.word);
 
 SENZ(tmp_cpu, tmp_byte1);

 tmp_cpu->cycles += 6 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* ASL A -> Arithmetic Shift Left on Accumulator */
INLINE void O_ASLA(void) {

 CLEAR_FLAG(tmp_cpu,FLAG_carry);
 if (tmp_cpu->A & 0x80) SET_FLAG(tmp_cpu,FLAG_carry);

 /* Do the bit shift */
 tmp_cpu->A <<= 1;

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* BCC -> Branch on Carry Clear */
INLINE void O_BCC(void) {

 if (!IS_FLAG_SET(tmp_cpu,FLAG_carry)) {
  tmp_cpu->PC.word = effective_address.word;
  tmp_cpu->cycles += cycle_modifier;
 } else tmp_cpu->PC.word++;

 tmp_cpu->cycles += 2;
}

/* BCS -> Branch on Carry Set */
INLINE void O_BCS(void) {

 if (IS_FLAG_SET(tmp_cpu,FLAG_carry)) {
  tmp_cpu->PC.word = effective_address.word;
  tmp_cpu->cycles += cycle_modifier;
 } else  tmp_cpu->PC.word++;

 tmp_cpu->cycles += 2;
}

/* BEQ -> Branch on Z EQual 1 */
INLINE void O_BEQ(void) {
 if (IS_FLAG_SET(tmp_cpu,FLAG_zero)) {
  tmp_cpu->PC.word = effective_address.word;
  tmp_cpu->cycles += cycle_modifier;
 } else  tmp_cpu->PC.word++;

 tmp_cpu->cycles += 2;
}

/* BIT -> test BITs */
INLINE void O_BIT(void) {
 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

 CLEAR_FLAG(tmp_cpu,FLAG_overflow);
 CLEAR_FLAG(tmp_cpu,FLAG_negative);
 CLEAR_FLAG(tmp_cpu,FLAG_zero);

 /* Put B7 of memory location in N */
 if (tmp_byte1 & 0x80) SET_FLAG(tmp_cpu,FLAG_negative);

 /* Put B6 of memory location in V */
 if (tmp_byte1 & 0x40) SET_FLAG(tmp_cpu,FLAG_overflow);

 /* If A and M = 0, set Z */
 if ((tmp_cpu->A & tmp_byte1)==0) SET_FLAG(tmp_cpu,FLAG_zero);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* BMI -> Branch on MInus result (N=1) */
INLINE void O_BMI(void) {

 if (IS_FLAG_SET(tmp_cpu,FLAG_negative)) {
  tmp_cpu->PC.word = effective_address.word;
  tmp_cpu->cycles += cycle_modifier;
 } else  tmp_cpu->PC.word++;

 tmp_cpu->cycles += 2;
}

/* BNE -> Branch on Z Not Equal 1 */
INLINE void O_BNE(void) {
 if (!IS_FLAG_SET(tmp_cpu,FLAG_zero)) {
  tmp_cpu->PC.word = effective_address.word;
  tmp_cpu->cycles += cycle_modifier;
 } else  tmp_cpu->PC.word++;

 tmp_cpu->cycles += 2;
}

/* BPL -> Branch on PLus result (N=0) */
INLINE void O_BPL(void) {

 if (!IS_FLAG_SET(tmp_cpu,FLAG_negative)) {
  tmp_cpu->PC.word = effective_address.word;
  tmp_cpu->cycles += cycle_modifier;
 } else  tmp_cpu->PC.word++;

 tmp_cpu->cycles += 2;
}

/* BRK -> BRK interrupt */
INLINE void O_BRK(void) {

 t6502_trigger_BRK(tmp_cpu);

}

/* BVC -> Branch on oVerflow Clear */
INLINE void O_BVC(void) {

 if (!IS_FLAG_SET(tmp_cpu,FLAG_overflow)) {
  tmp_cpu->PC.word = effective_address.word;
  tmp_cpu->cycles += cycle_modifier;
 } else  tmp_cpu->PC.word++;

 tmp_cpu->cycles += 2;
}

/* BVS -> Branch on oVerflow Set */
INLINE void O_BVS(void) {

 if (IS_FLAG_SET(tmp_cpu,FLAG_overflow)) {
  tmp_cpu->PC.word = effective_address.word;
  tmp_cpu->cycles += cycle_modifier;
 } else  tmp_cpu->PC.word++;

 tmp_cpu->cycles += 2;
}

/* CLC -> CLear Carry flag */
INLINE void O_CLC(void) {
 CLEAR_FLAG(tmp_cpu,FLAG_carry);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* CLD -> CLear Decimal flag */
INLINE void O_CLD(void) {
 CLEAR_FLAG(tmp_cpu,FLAG_decimal);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* CLI -> CLear Interrupt disable flag */
INLINE void O_CLI(void) {
 CLEAR_FLAG(tmp_cpu,FLAG_interrupt);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++; 
}

/* CLV -> CLear oVerflow flag */
INLINE void O_CLV(void) {
 CLEAR_FLAG(tmp_cpu,FLAG_overflow);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++; 
}

/* CMP -> CoMPare A with memory */

INLINE void O_CMP(void) {
 /* Get the value to compare */
 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

/* if ((tmp_cpu->A + 0x100 - tmp_byte1) > 0xff)
  SET_FLAG(tmp_cpu, FLAG_carry);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_carry);
  
 tmp_byte1 = (BYTE)(tmp_cpu->A + 0x100 - tmp_byte1);
 if (tmp_byte1)
  CLEAR_FLAG(tmp_cpu, FLAG_zero);
 else
  SET_FLAG(tmp_cpu, FLAG_zero);
  
 if (tmp_byte1 & 0x80)
  SET_FLAG(tmp_cpu, FLAG_negative);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_negative);
 */

 if (tmp_cpu->A + 0x100 - tmp_byte1 > 0xff)
  SET_FLAG(tmp_cpu, FLAG_carry);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_carry);

 tmp_byte1 = tmp_cpu->A + 0x100 - tmp_byte1;
 SENZ(tmp_cpu,tmp_byte1);
 
 
 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* CPX -> CoMPare X with memory */
INLINE void O_CPX(void) {
 /* Get the value to compare */
 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

/* if ((tmp_cpu->X + 0x100 - tmp_byte1) > 0xff)
  SET_FLAG(tmp_cpu, FLAG_carry);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_carry);
  
 tmp_byte1 = (BYTE)(tmp_cpu->X + 0x100 - tmp_byte1);
 if (tmp_byte1)
  CLEAR_FLAG(tmp_cpu, FLAG_zero);
 else
  SET_FLAG(tmp_cpu, FLAG_zero);
  
 if (tmp_byte1 & 0x80)
  SET_FLAG(tmp_cpu, FLAG_negative);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_negative);
 */

 if (tmp_cpu->X + 0x100 - tmp_byte1 > 0xff)
  SET_FLAG(tmp_cpu, FLAG_carry);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_carry);

 tmp_byte1 = tmp_cpu->X + 0x100 - tmp_byte1;
 SENZ(tmp_cpu,tmp_byte1);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* CPY -> CoMPare Y with memory */
INLINE void O_CPY(void) {
 /* Get the value to compare */
 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

/* if ((tmp_cpu->Y + 0x100 - tmp_byte1) > 0xff)
  SET_FLAG(tmp_cpu, FLAG_carry);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_carry);
  
 tmp_byte1 = (BYTE)(tmp_cpu->Y + 0x100 - tmp_byte1);
 if (tmp_byte1)
  CLEAR_FLAG(tmp_cpu, FLAG_zero);
 else
  SET_FLAG(tmp_cpu, FLAG_zero);
  
 if (tmp_byte1 & 0x80)
  SET_FLAG(tmp_cpu, FLAG_negative);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_negative);
 */

 if (tmp_cpu->Y + 0x100 - tmp_byte1 > 0xff)
  SET_FLAG(tmp_cpu, FLAG_carry);
 else
  CLEAR_FLAG(tmp_cpu, FLAG_carry);

 tmp_byte1 = tmp_cpu->Y + 0x100 - tmp_byte1;
 SENZ(tmp_cpu,tmp_byte1);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/**********************************************************************/
/* DEC -> DECrease memory */
INLINE void O_DEC(void) {

 /* Read from address */
 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);
 /* Decrease value at memory */
 tmp_byte1--;
 /* Write back to memory */
 tmp_cpu->write_mem(tmp_byte1, effective_address.word);

 SENZ(tmp_cpu, tmp_byte1);

 tmp_cpu->cycles += 6 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* DEX -> DEcrease X register */
INLINE void O_DEX(void) {
 /* Decrease register */
 tmp_cpu->X--;

  SENZ(tmp_cpu, tmp_cpu->X);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++; 
}

/* DEY -> DEcrease Y register */
INLINE void O_DEY(void) {
 /* Decrease register */
 tmp_cpu->Y--;

 SENZ(tmp_cpu, tmp_cpu->Y);
 
 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++; 
}

/* EOR -> Exclusive OR on memory */
INLINE void O_EOR(void) {

 tmp_cpu->A ^= tmp_cpu->read_mem(effective_address.word);

 SENZ(tmp_cpu, tmp_cpu->A);
 
 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* INC -> INCrease memory */
INLINE void O_INC(void) {

 /* Read from address */
 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);
 /* Decrease value at memory */
 tmp_byte1++;
 /* Write back to memory */
 tmp_cpu->write_mem(tmp_byte1, effective_address.word);

 SENZ(tmp_cpu, tmp_byte1);
 
 tmp_cpu->cycles += 6 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* INX -> INcrease X register */
INLINE void O_INX(void) {
 /* Decrease register */
 tmp_cpu->X++;

 SENZ(tmp_cpu, tmp_cpu->X);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++; 
}

/* INY -> INcrease Y register */
INLINE void O_INY(void) {
 /* Decrease register */
 tmp_cpu->Y++;

 SENZ(tmp_cpu, tmp_cpu->Y);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++; 
}


/* JMP -> JuMP to address */
INLINE void O_JMP(void) {

 /* Put effective address in Program Counter */
 tmp_cpu->PC.word = effective_address.word;

 tmp_cpu->cycles += 3 + cycle_modifier;
}

/* JSR -> Jump to Sub Routine */
INLINE void O_JSR(void) {

 /* Push old PC on stack */
 PUSH_BYTE(tmp_cpu, tmp_cpu->PC.byte.h);
 PUSH_BYTE(tmp_cpu, tmp_cpu->PC.byte.l);

 /* Put effective address in Program Counter */
 tmp_cpu->PC.word = effective_address.word;

 tmp_cpu->cycles += 6;
}

/* LDA -> LoaD memory to A */
INLINE void O_LDA(void) {

 /* Put contents of effective address in register */
 tmp_cpu->A = tmp_cpu->read_mem(effective_address.word);

 SENZ(tmp_cpu, tmp_cpu->A);
 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* LDX -> LoaD memory to X */
INLINE void O_LDX(void) {

 /* Put contents of effective address in register */
 tmp_cpu->X = tmp_cpu->read_mem(effective_address.word);

 SENZ(tmp_cpu, tmp_cpu->X);
 
 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* LDY -> LoaD memory to Y */
INLINE void O_LDY(void) {

 /* Put contents of effective address in register */
 tmp_cpu->Y = tmp_cpu->read_mem(effective_address.word);

 SENZ(tmp_cpu, tmp_cpu->Y);
 
 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* LSR -> Logical Shift Right */
INLINE void O_LSR(void) {

 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

 CLEAR_FLAG(tmp_cpu,FLAG_negative | FLAG_carry);

 if (tmp_byte1 & 0x01) SET_FLAG(tmp_cpu,FLAG_carry);

 /* Do the bit shift */
 tmp_byte1 >>= 1;

 /* Write back to memory */
 tmp_cpu->write_mem(tmp_byte1, effective_address.word);
 
 /* Set ZERO flag */
 if (tmp_byte1)
  CLEAR_FLAG(tmp_cpu,FLAG_zero);
 else
  SET_FLAG(tmp_cpu,FLAG_zero);

 tmp_cpu->cycles += 6 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* LSR A -> Logical Shift Right Accumulator */
INLINE void O_LSRA(void) {

 CLEAR_FLAG(tmp_cpu,FLAG_negative);
 CLEAR_FLAG(tmp_cpu,FLAG_carry);
 if (tmp_cpu->A & 0x01) SET_FLAG(tmp_cpu,FLAG_carry);

 /* Do the bit shift */
 tmp_cpu->A >>= 1;

 /* Set ZERO flag */
 if (tmp_cpu->A)
  CLEAR_FLAG(tmp_cpu,FLAG_zero);
 else
  SET_FLAG(tmp_cpu,FLAG_zero);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* NOP -> No OPeration */
INLINE void O_NOP(void) {
 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* ORA -> OR on Accumulator */
INLINE void O_ORA(void) {

 tmp_cpu->A |= tmp_cpu->read_mem(effective_address.word);

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* PHA -> PusH A on the stack */
INLINE void O_PHA(void) {
 PUSH_BYTE(tmp_cpu,tmp_cpu->A);

 tmp_cpu->cycles += 3;
 tmp_cpu->PC.word++;
}

/* PHP -> PusH P on the stack */
INLINE void O_PHP(void) {
 PUSH_BYTE(tmp_cpu,tmp_cpu->status);

 tmp_cpu->cycles += 3;
 tmp_cpu->PC.word++;
}

/* PLA -> PuLl A from the stack */
INLINE void O_PLA(void) {
 POP_BYTE(tmp_cpu,tmp_cpu->A);

 tmp_cpu->cycles += 4;
 tmp_cpu->PC.word++;
}

/* PLP -> PuLl P from the stack */
INLINE void O_PLP(void) {
 POP_BYTE(tmp_cpu,tmp_cpu->status);

 tmp_cpu->cycles += 4;
 tmp_cpu->PC.word++;
}

/* ROL -> ROtate Left */
INLINE void O_ROL(void) {

 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

 tmp_byte2 = (tmp_byte1 & 0x80) >> 7;
 if (tmp_byte2) SET_FLAG(tmp_cpu,FLAG_carry); else CLEAR_FLAG(tmp_cpu,FLAG_carry);
 tmp_byte1 <<= 1;
 tmp_byte1 |= tmp_byte2;

 SENZ(tmp_cpu, tmp_byte1);

 tmp_cpu->write_mem(tmp_byte1, effective_address.word);
 
 tmp_cpu->cycles += 6 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* ROL A -> ROtate Left Accumulator */
INLINE void O_ROLA(void) {

/* tmp_cpu->A = tmp_cpu->read_mem(effective_address.word); */

 tmp_byte2 = (tmp_cpu->A & 0x80) >> 7;
 if (tmp_byte2) SET_FLAG(tmp_cpu,FLAG_carry); else CLEAR_FLAG(tmp_cpu,FLAG_carry);
 tmp_cpu->A <<= 1;
 tmp_cpu->A |= tmp_byte2;

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++; 
}

/* ROR -> ROtate Right */
INLINE void O_ROR(void) {

 tmp_byte1 = tmp_cpu->read_mem(effective_address.word);

 tmp_byte2 = tmp_byte1 & 0x01;
 if (tmp_byte2) SET_FLAG(tmp_cpu,FLAG_carry); else CLEAR_FLAG(tmp_cpu,FLAG_carry);
 tmp_byte1 >>= 1;
 tmp_byte1 |= tmp_byte2 << 7;

 SENZ(tmp_cpu, tmp_byte1);

 tmp_cpu->write_mem(tmp_byte1, effective_address.word);
 
 tmp_cpu->cycles += 6 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* ROR A -> ROtate Right Accumulator */
INLINE void O_RORA(void) {

 /* tmp_cpu->A = tmp_cpu->read_mem(effective_address.word); */

 tmp_byte2 = tmp_cpu->A & 0x01;
 if (tmp_byte2) SET_FLAG(tmp_cpu,FLAG_carry); else CLEAR_FLAG(tmp_cpu,FLAG_carry);
 tmp_cpu->A >>= 1;
 tmp_cpu->A |= tmp_byte2 << 7;

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++; 
}

/* RTI -> ReTurn from Interrupt */
INLINE void O_RTI(void) {

 /* Pop P */
 POP_BYTE(tmp_cpu, tmp_cpu->status);
 /* Pop PC */
 POP_BYTE(tmp_cpu, tmp_cpu->PC.byte.l);
 POP_BYTE(tmp_cpu, tmp_cpu->PC.byte.h);

 tmp_cpu->int_nesting--;
 tmp_cpu->cycles += 6;
}

/* RTS -> ReTurn from Subroutine */
INLINE void O_RTS(void) {

 /* Pop PC */
 POP_BYTE(tmp_cpu, tmp_cpu->PC.byte.l);
 POP_BYTE(tmp_cpu, tmp_cpu->PC.byte.h);

 tmp_cpu->PC.word++;

 tmp_cpu->cycles += 6;
}

/* SBC -> SuBstract with Carry */
INLINE void O_SBC(void) {
 int tmp_word3;

 tmp_byte1 = tmp_cpu->read_mem(effective_address.word) ^ 0xff;

 carry_save = (tmp_cpu->status & 0x01);
 
 tmp_word3 = (BYTE)tmp_cpu->A + (BYTE)tmp_byte1  + (carry_save << 4);

 if ((tmp_word3 > 127) || (tmp_word3 < -128))
  SET_FLAG(tmp_cpu,FLAG_overflow);
 else
  CLEAR_FLAG(tmp_cpu,FLAG_overflow);
  
 tmp_word3 = tmp_cpu->A + tmp_byte1 + carry_save;

 if (tmp_word3 > 255)
  SET_FLAG(tmp_cpu,FLAG_carry);
 else
  CLEAR_FLAG(tmp_cpu,FLAG_carry);

 tmp_cpu->A = tmp_word3 & 0xff;

 if (IS_FLAG_SET(tmp_cpu,FLAG_decimal)) {
  tmp_cpu->A -= 0x66;
  CLEAR_FLAG(tmp_cpu,FLAG_carry);

  if ((tmp_cpu->A & 0x0f) > 0x09) tmp_cpu->A += 0x06;

  if ((tmp_cpu->A & 0xf0) > 0x90) {

   tmp_cpu->A += 0x60;
   SET_FLAG(tmp_cpu,FLAG_carry);
  }
 }

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++; 
}

/* SEC -> SEt Carry flag */
INLINE void O_SEC(void) {
 SET_FLAG(tmp_cpu,FLAG_carry);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* SEI -> SEt Interrupt disable flag */
INLINE void O_SED(void) {
 SET_FLAG(tmp_cpu,FLAG_decimal);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* SEI -> SEt Interrupt disable flag */
INLINE void O_SEI(void) {
 SET_FLAG(tmp_cpu,FLAG_interrupt);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* STA -> STore Accumulator to memory */
INLINE void O_STA(void) {

 tmp_cpu->write_mem(tmp_cpu->A, effective_address.word);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* STX -> STore X register to memory */
INLINE void O_STX(void) {

 tmp_cpu->write_mem(tmp_cpu->X, effective_address.word);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* STY -> STore Y register to memory */
INLINE void O_STY(void) {

 tmp_cpu->write_mem(tmp_cpu->Y, effective_address.word);

 tmp_cpu->cycles += 4 + cycle_modifier;
 tmp_cpu->PC.word++;
}

/* TAX -> Transfer A to X */
INLINE void O_TAX(void) {

 tmp_cpu->X = tmp_cpu->A;

 SENZ(tmp_cpu, tmp_cpu->X);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* TAY -> Transfer A to Y */
INLINE void O_TAY(void) {

 tmp_cpu->Y = tmp_cpu->A;

 SENZ(tmp_cpu, tmp_cpu->Y);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* TSX -> Transfer S to X */
INLINE void O_TSX(void) {

 tmp_cpu->X = tmp_cpu->S;

 SENZ(tmp_cpu, tmp_cpu->X);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* TXA -> Transfer X to A */
INLINE void O_TXA(void) {

 tmp_cpu->A = tmp_cpu->X;

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* TYA -> Transfer Y to A */
INLINE void O_TYA(void) {

 tmp_cpu->A = tmp_cpu->Y;

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

/* TXS -> Transfer X to S */
INLINE void O_TXS(void) {

 tmp_cpu->S = tmp_cpu->X;

 SENZ(tmp_cpu, tmp_cpu->A);

 tmp_cpu->cycles += 2;
 tmp_cpu->PC.word++;
}

INLINE void O_BADOP(void) {
 fprintf(stderr,"Invalid opcode: [$%02X] at PC=$%04X\n",opcode,tmp_cpu->PC.word);
 tmp_cpu->PC.word++; 
}

void t6502_run(CPU_6502_unit *CPU, unsigned long c) {
  tmp_cpu = CPU;

 while (CPU->cycles <= c) {
  opcode = CPU->read_op(CPU->PC.word);

  switch(opcode) {
#include "optable.h"   /*OPtable;*/
  }
 }
}

void t6502_run_step(CPU_6502_unit *CPU) {
  tmp_cpu = CPU;

  opcode = CPU->read_op(CPU->PC.word);

  switch(opcode) {
#include "optable.h" /*OPtable;*/
  }
}


void t6502_init_emu()
{
}


