/*************************************************************************/
/* T6502 -> Tennessee's 6502 CPU core [t6502.h: Header file]             */
/*                                                                       */
/*************************************************************************/

#include <stdio.h>

/* ifndef block to be sure that we don't re-define stuff */
#ifndef T6502_HEADER
#define T6502_HEADER

#ifdef __GNUC__
#define INLINE inline
#else
#define INLINE static
#endif

/* Base address of the stack */
#define STACK_BASE 0x100

#define LSB_FIRST

#define FLAG_carry      1 /* Carry flag */
#define FLAG_zero       2 /* ZERO flag */
#define FLAG_interrupt  4 /* Interrupt disable flag */
#define FLAG_decimal    8 /* Decimal flag */
#define FLAG_BRK        16 /* Software interrupt in progress flag */
#define FLAG_unused     32 /* Unused flag, always 1 */
#define FLAG_overflow   64 /* Arithmetic overflow flag */
#define FLAG_negative   128 /* Negative sign flag */

/* Macro to clear a flag */
#define CLEAR_FLAG(p,n) p->status &= ~n
/* Macro to set a flag */
#define SET_FLAG(p,n) p->status |= n
/* Macro to check if a flag is set */
#define IS_FLAG_SET(p,n) (p->status & n) ? 1 : 0

/* Macro to PUSH something in the stack */
#define PUSH_BYTE(p,d) p->write_mem(d,(STACK_BASE | p->S)); p->S--
/* Macro to POP something from the stack */
#define POP_BYTE(p,d) p->S++; d = p->read_mem(STACK_BASE | p->S)

typedef unsigned char BYTE; /* 8-bits byte */
typedef unsigned short WORD; /* 16-bits WORD */
typedef signed char toffset; /* 7-bits + sign for Relative addressing offsets */

/* Taken from Marat Fayzullin's M6502 */
typedef union
{
#ifdef LSB_FIRST
  struct { BYTE l,h; } byte;
#else
  struct { BYTE h,l; } byte;
#endif
  WORD word;
} tpair;

typedef struct
{
  char name[17]; /* CPU's name/description, 16 chars max */
  
  BYTE A; /* Accumulator */

  /* X and Y index registers */
  BYTE X;
  BYTE Y;

  BYTE S; /* Stack Pointer */

  BYTE status; /* Status register (flags) */

  tpair PC; /* Program Counter */

  unsigned long cycles; /* Cycle counter */

  BYTE (*read_mem)(WORD address);
  BYTE (*read_op)(WORD address);
  void (*write_mem)(BYTE data, WORD address);

  int int_nesting; /* Interrupt nesting control */

} CPU_6502_unit;

extern BYTE cycles[256]; /* Array of cycles */

void t6502_init_cpu(CPU_6502_unit *CPU, BYTE (*read_func)(WORD A), BYTE (*read_op_func)(WORD A), void (*write_func)(BYTE d, WORD A), char *base_name );
void t6502_trigger_IRQ(CPU_6502_unit *CPU);
void t6502_trigger_NMI(CPU_6502_unit *CPU);
void t6502_trigger_BRK(CPU_6502_unit *CPU);
void t6502_trigger_RESET(CPU_6502_unit *CPU);
void t6502_run(CPU_6502_unit *CPU, unsigned long c);
void t6502_run_step(CPU_6502_unit *CPU);
void t6502_init_emu(void);

#endif
